/*

Basic code:    BCM2835 "GPU_FFT" release 3.0	   Copyright (c) 2015, Andrew Holme
Modifications: Discrete Wavelet Transform example, Copyright (c) 2017, Piotr Romaniuk, ELESOFTROM

All rights reserved

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the copyright holder nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <string.h>
#include <stdio.h>

#include "gpu_fft.h"
#include "gpu_fft_shader.h"
#include "gpu_fft_base.h"

#include "gpu_dwt.h"

#define GPU_BUSY_WAIT_LIMIT (5<<12) // ~1ms


int gpu_prepare( int mb, int log2_N, int jobs, struct gpu_memory_dsc *gpu_dsc, int n_simd16) {

    unsigned int code_bytes;
    unsigned int size;
    int i, q, ret;

    struct GPU_FFT_BASE *base;

    code_bytes = gpu_fft_shader_size(log2_N);

    size = sizeof(struct gpu_memory_image);//TODO add dynamic size, growing ping/pong buffers

    printf("mem_alloc::size=%d, PTR: ", size );
    //allocate common memory for communication ARM->VC->ARM
    ret = gpu_alloc(mb, size, gpu_dsc);
    if (ret) return ret;

    printf("ptr.vc=%08x, .asm.vptr=%p\n", gpu_dsc->vc_ptr, gpu_dsc->image );
    // Prepare and fill up image in allocated memory block:
    // Header
    base = &gpu_dsc->image->info.base;

    PRINT_PTR("PING-PONG ", gpu_dsc, data );
    PRINT_PTR("DBG-rfiles", gpu_dsc, dbg.rfile );
    PRINT_PTR("DBG-accus ", gpu_dsc, dbg.accus );
    PRINT_PTR("CODE      ", gpu_dsc, code);

    // Copy shader code to image
    memcpy( gpu_dsc->image->code, gpu_fft_shader_code(log2_N), code_bytes);
    base->vc_code = vcptr_of_field( gpu_dsc, code);

    PRINT_PTR("Filter    ", gpu_dsc, filter);
    printf("\n");

    for (q=0; q<GPU_FFT_QPUS; q++) {
    	struct qpu_uniforms *uni = &gpu_dsc->image->uniforms[q];
    	uni->dbg_rfiles = vcptr_of_field( gpu_dsc, dbg.rfile );
    	uni->dbg_accus  = vcptr_of_field( gpu_dsc, dbg.accus );
    	printf("UNIF.QPU%d: dbg_rfiles=%08x, accus=%08x\n", q, uni->dbg_rfiles, uni->dbg_accus);

    	uni->filter = vcptr_of_field( gpu_dsc, filter );
    	printf("UNIF.QPU%d: filter=%08x\n", q, uni->filter );
        uni->qpu_number = QPU_NUMBER;
        printf("UNIF.QPU%d: qpu_id\n", q);
        uni->qpu_id = q;  //QPU id (0=Master, 1-7 slaves)
        uni->v16_count_per_qpu = n_simd16;//length of the input data [in 16 samples blocks]
        // for each job, specify ping and pong pair of buffers:
        for (i=0; i<jobs; i++) {
        	uni->in_data = vcptr_of_field( gpu_dsc, data.ping_buffer );//*uptr++ = vc_data + data_bytes*i;
            uni->out_data = vcptr_of_field( gpu_dsc, data.pong_buffer );//TODO add jobs functionality; *uptr++ = vc_data + data_bytes*i + data_bytes*jobs;
        	printf("UNIF.QPU%d: jobs-spec1=%08x ~2=%08x\n", q, uni->in_data, uni->out_data );
        }
        // just to mark end of jobs for QPU loop write zero address and IRQ enable flag
        // The IRQ flag is only set for master (q=0)
        //printf("UNIF.QPU%d: last: ~1=%08x ~2=%08x\n", q, 0, (q==0));
  //      *uptr++ = 0;
  //      *uptr++ = (q==0); // For mailbox: IRQ enable, master only

        base->vc_unifs[q] = vcptr_of_field( gpu_dsc, uniforms[q]);//  gpu_fft_ptr_inc(&ptr, sizeof(int)*(1+5+jobs*2));
        printf("base->unifs[%d] = %08x\n", q, base->vc_unifs[q]);
    }
    int j, k;
    unsigned int * uniforms = (unsigned int*)&gpu_dsc->image->uniforms;

    printf("Uniforms\n");
    for(k=0;k<QPU_NUMBER;k++){
     printf("\n");
     for(j=0;j<8;j++){
    	char* dsc[]={"d-regs","d-acc","filtr","qpu_numb","qpuid","count","ping(in)","pong(out)"};
		printf("%p [%d]: %08x %s\n", uniforms+j+k*8, j, uniforms[j+k*8], dsc[j] );
     }
    }
    PRINT_PTR("UNIFORMS", gpu_dsc, uniforms );

    if ((jobs<<log2_N) <= GPU_BUSY_WAIT_LIMIT) {
        // execution by direct register poking with busy wait for end
        base->vc_msg = 0;
    }
    else {
        // execution via mailbox message
        for (q=0; q<GPU_FFT_QPUS; q++) {
            gpu_dsc->image->mbox_exe[q].uniforms = base->vc_unifs[q];
            gpu_dsc->image->mbox_exe[q].code = base->vc_code;
        }
        base->vc_msg = vcptr_of_field( gpu_dsc, mbox_exe );
    }


    return 0;
}


unsigned gpu_execute(struct GPU_FFT_BASE *base) {
    return gpu_fft_base_exec(base, QPU_NUMBER);
}

void gpu_release(struct GPU_FFT_BASE *base) {
    gpu_fft_base_release(base);
}

